module Test where

import Data.Maybe(isJust,fromJust)
import Data.List(nub)

import Test.Tasty
import Test.Tasty.HUnit
import qualified Test.Tasty.QuickCheck as QC
import qualified Test.Tasty.SmallCheck as SC

import Delivery

type TG = Graph String

--- The simple graph from the exercise sheet
simple :: TG
simple = a where
  a = N "A" [E 1 b, E 1 c]
  b = N "B" [E 1 a, E 1 e, E 9 d]
  c = N "C" [E 1 a, E 1 d, E 9 e]
  d = N "D" [E 9 b, E 1 c, E 1 f]
  e = N "E" [E 1 b, E 9 c, E 1 f]
  f = N "F" [E 1 d, E 1 e]
  
--- A simple graph which is not disconnected but has no Hamiltonian cycle
noHC :: TG
noHC = a where
  a = N "A" [E 1 b, E 2 c, E 2 d]
  b = N "B" [E 1 a]
  c = N "C" [E 2 a, E 3 d]
  d = N "D" [E 2 a, E 3 c]

-- Auxiliary function: check wether something *is* a path
isPath :: Eq n=> Node n-> Path n-> Bool
isPath n (P w ps) = isP n w ps where
  isP _ weight [] = False
  isP (N n _) weight  (p:[]) = n == p && weight == 0
  isP (N n es) weight (p:qs) = n == p && goto es where
    goto (E w t: es) = isP t (weight- w) qs || goto es 
    goto [] = False

-- A cycle is a non-empty path where the last node is equal to the first one,
-- and in the rest there is repetition repetition repetition
isCycle :: Eq n=> Path n-> Bool
isCycle (P _ xs) = not (null xs) && head xs == last xs && noRepeat (tail xs)

noRepeat :: Eq n=> [n]-> Bool
noRepeat xs = length (nub xs) == length xs

-- Tests

run = defaultMain $ testGroup "All tests" [pathTests, cycleTests, nodeTests, tppTests]

pathTests = testGroup "Tests for Path"
  [ testCase "Adding a node" $
    add (P 17 "abc") 3 'x' @=? P 20 "xabc"
  , testCase "Path contains node" $
    contains (P 17 "abcde") 'c' @? "Does not"
  , testCase "Path does not contain node" $
    not (contains (P 17 "abcde") 'x') @? "It does"
  ]

cycleTests = testGroup "Tests for cycle"
  [ testCase "counterexample has 6 cycles" $
    length (cycles noHC) @=? 6
  , testCase "simple graph has  11 cycles" $
    length (cycles simple) @=? 11
  , testCase "cycles returns only cycles (simple graph)" $
    all isCycle (cycles simple) @? "Some are not"
  , testCase "All cycles are path (simple graph)" $
    all (isPath simple) (cycles simple) @? "Some are not"
  , testCase "All cycles are path (counterexample)" $
    all (isPath noHC) (cycles noHC) @? "Some are not paths"
  ]

nodeTests = testGroup "Test for node function"
  [ testCase "simple graph has 6 nodes" $
    numNodes simple @=? 6 
  , testCase "counterexample graph has 4 nodes" $
    numNodes noHC @=? 4
  ]

tppTests = testGroup "Test for travelling pizza problem"
  [ testCase "counterexample has no solution" $
    tpp noHC @=? Nothing
  , testCase "simple graph has a shortest H-cycle" $
    isJust (tpp simple) @? "Is None"
  , testCase "simple graph shortest H-cycle is a path" $
    isPath simple (fromJust (tpp simple)) @? "Not a path"
  , testCase "simple graph shortest H-cycle is a cycle" $
    isCycle (fromJust (tpp simple)) @? "Not a cycle"
  , testCase "simple graph shortest H-cycle has length 7" $
    length (path (fromJust (tpp simple))) @=? 7
  , testCase "simple graph shortest H-cycle has cost 6" $
    cost (fromJust (tpp simple)) @=? 6
  ]


-- This is Bremen, as from the exercise sheet.
-- Use this graph to make your own tests (do not change the graph, though).
bremen :: TG
bremen = b1 where
  b1 = N "Viertel" [E 5 b2, E 5 b3, E 7 b4, E 6 b12, E 7 b10]
  b2 = N "Innenstadt" [E 5 b1, E 8 b3, E 6 b6, E 3 b7, E 9 b9]
  b3 = N "Schwachhausen" [E 5 b1, E 8 b2, E 7 b6, E 4 b4, E 4 b5]
  b4 = N "Horn" [E 4 b3, E 7 b1, E 2 b5, E 5 b12]
  b5 = N "Uni" [E 4 b3, E 2 b4, E 6 b6, E 3 b8]
  b6 = N "Findorff" [E 6 b2, E 7 b3, E 6 b5, E 5 b7]
  b7 = N "Walle" [E 3 b2, E 5 b6, E 4 b8]
  b8 = N "Gröpelingen" [E 3 b5, E 4 b7]
  b9 = N "Neustadt" [E 9 b2, E 5 b11]
  b10 = N "Habenhausen" [E 7 b1, E 4 b11]
  b11 = N "Kattenturm" [E 5 b9, E 4 b10, E 7 b12]
  b12 = N "Hemelingen" [E 5 b1, E 5 b4, E 7 b11]
